(* This is the source calculus of the sample client. It is a core ML. *)

(* Nominal representation of term variables and binders. *)

(* BEGIN ML *)
type tevar = string           
type type_id = Type of string
type label_id = Label of string

type ('a,'b) typ =
  | TyVar of 'a
  | TyArrow of ('a,'b) typ * ('a,'b) typ
  | TyProduct of ('a,'b) typ list
  | TyConstr of type_id * ('a,'b) typ list

type ('a, 'b) annotation = 'b list * ('a, 'b) typ
  (* some <flexible vars> . <ty> *)

type nominal_type = (string,string) typ
type nominal_annotation = (string,string) annotation

type term =
  | Var of tevar
  | Abs of tevar * term
  | App of term * term
  | Let of tevar * term * term
(* END ML *)
  | Annot of term * nominal_annotation
  | Tuple of term list
  | LetProd of tevar list * term * term
  | Variant of label_id * term
  | Match of term * branch list

and branch = pattern * term

and pattern =
  | PVar of tevar
  | PWildcard
  | PAnnot of pattern * nominal_annotation
  | PVariant of label_id * pattern
  | PTuple of pattern list
                  
type data_type_descr = {
    name : type_id ;
    type_params : string list ;
    data_kind : bool ; (* true for variants, false for records *)
    labels_descr : label_descr list
  }

and label_descr = {
    label_name : label_id ;
    type_name : type_id ;
    arg_type : nominal_type
  }
