open Client

(* A few manually constructed terms. *)

let x =
  ML.Var "x"

let y =
  ML.Var "y"

let id =
  ML.Abs ("x", x)

let delta =
  ML.Abs ("x", ML.App (x, x))

let deltadelta =
  ML.App (delta, delta)

let idid =
  ML.App (id, id)

let k =
  ML.Abs ("x", ML.Abs ("y", x))

let genid =
  ML.Let ("x", id, x)

let genidid =
  ML.Let ("x", id, ML.App (x, x))

let genkidid =
  ML.Let ("x", ML.App (k, id), ML.App (x, id))

let genkidid2 =
  ML.Let ("x", ML.App (ML.App (k, id), id), x)

let app_pair = (* ill-typed *)
  ML.App (ML.Tuple [id; id], id)

(* option *)
let option_env =
  
  (* type 'a option = None | Some of 'a *)
  let option_typedecl = ML.{
    name = Type "option";
    type_params = [ "a" ];
    data_kind = true;
    labels_descr = [
      {
        label_name = Label"None";
        type_name = Type "option";
        arg_type = TyProduct [];          
      } ; {
        label_name = Label "Some";
        type_name = Type "option";
        arg_type = TyVar "a";
      }
    ]
  } in

  Infer.Env.add_typedecl Infer.Env.empty option_typedecl
  
let none = ML.Variant ( ML.Label "None" , ML.Tuple [] )

let some =
  ML.Variant (
    ML.Label "Some",
    id
  )

let match_none = ML.(
    Match (none, [
      PVariant (Label "None", PTuple []), none ;
      PVariant (Label "Some", PTuple [(PVar "x")]), x
    ])
  )

let match_some = ML.(
    Match (some, [
      PVariant (Label "None", PTuple []), none ;
      PVariant (Label "Some", PWildcard), none
    ])
  )

let match_some_annotated = ML.(
    Match (some, [
      ( PVariant (Label "None", PTuple []), none );
      ( PAnnot (PVariant (Label "Some", PWildcard),
               (["a"], TyConstr (Type "option", [ TyVar "a" ]))),
        none );
    ])
  )

(* list *)
let list_env =

  (* type 'a list = Nil | Cons of 'a * 'a list *)
  let list_typedecl = ML.{
    name = Type "list";
    type_params = [ "a" ];
    data_kind = true;
    labels_descr = [
      {
        label_name = Label "Nil";
        type_name = Type "list";
        arg_type = TyProduct [];          
      } ; {
        label_name = Label "Cons";
        type_name = Type "list";
        arg_type = TyProduct [ TyVar "a" ; TyConstr (Type "list", [ TyVar "a" ]) ];
      }
    ]
  } in

  Infer.Env.add_typedecl option_env list_typedecl

let nil = ML.Variant ( ML.Label "Nil" , ML.Tuple [] )

let list =
  ML.Variant (
    ML.Label "Cons",
    ML.Tuple [ id ; nil ]
  )

let list_annotated =
  let open ML in
  Annot (
    Variant (
      Label "Cons",
      Tuple [ Annot (id, (["a"], TyArrow (TyVar "a", TyVar "a")));
              nil ]
    ),
    (["a"; "b"], TyConstr (Type "list", [TyArrow (TyVar "a", TyVar "b")]))
  )

(* tree *)
let tree_env =

  (* type 'a tree = Leaf | Node of 'a tree * 'a * 'a tree *)
  let tree_typedecl = ML.{
    name = Type "tree";
    type_params = [ "a" ];
    data_kind = true;
    labels_descr = [
      {
        label_name = Label "Leaf";
        type_name = Type "tree";
        arg_type = TyProduct []
      } ; {
        label_name = Label "Node";
        type_name = Type "tree";
        arg_type = TyProduct [
          TyConstr (Type "tree", [ TyVar "a" ]);
          TyVar "a";
          TyConstr (Type "tree", [ TyVar "a" ]);
        ]
      }
    ]
  } in

  Infer.Env.add_typedecl list_env tree_typedecl

let leaf = ML.Variant ( ML.Label "Leaf" , ML.Tuple [] )

let node =
  ML.Variant (
    ML.Label "Node",
    ML.Tuple [
      leaf ;
      id ;
      leaf ;
    ]
  )

let foo_env =

  let toto_typedecl = ML.{
    name = Type "toto";
    type_params = [];
    data_kind = true;
    labels_descr = [
      {
        label_name = Label "Foo";
        type_name = Type "toto";
        arg_type = TyProduct []
      } ; {
        label_name = Label "Toto";
        type_name = Type "toto";
        arg_type = TyProduct []
      }
    ]
  } in

  let toto_env = Infer.Env.add_typedecl tree_env toto_typedecl in

  let tata_typedecl = ML.{
    name = Type "tata";
    type_params = [];
    data_kind = true;
    labels_descr = [
      {
        label_name = Label "Foo";
        type_name = Type "toto";
        arg_type = TyProduct []
      } ; {
        label_name = Label "Tata";
        type_name = Type "toto";
        arg_type = TyProduct []
      }
    ]
  } in

  Infer.Env.add_typedecl toto_env tata_typedecl

let foo_something_list something =
  ML.Variant (
    ML.Label "Cons",
    ML.Tuple [
      ML.Variant (ML.Label "Foo", ML.Tuple []) ;
      ML.Variant (
        ML.Label "Cons",
        ML.Tuple [ something ;  ]
        )
      ]
    )

let toto = ML.Variant ( ML.Label "Toto", ML.Tuple [] )
let tata = ML.Variant ( ML.Label "Tata", ML.Tuple [] )

let foo_toto_list = foo_something_list toto
let foo_tata_list = foo_something_list tata
  

let () = (*
  assert (Test.CheckML.test Infer.Env.empty idid);
  assert (Test.CheckML.test Infer.Env.empty genid);
  assert (Test.CheckML.test Infer.Env.empty genidid);
  assert (Test.CheckML.test Infer.Env.empty genkidid);
  assert (Test.CheckML.test Infer.Env.empty genkidid2);
  assert (Test.CheckML.test option_env none);
  assert (Test.CheckML.test option_env some);
  assert (Test.CheckML.test list_env nil);
  assert (Test.CheckML.test list_env list); *)
  assert (Test.CheckML.test list_env list_annotated); (*
  assert (Test.CheckML.test tree_env leaf);
  assert (Test.CheckML.test tree_env node);
  assert (Test.CheckML.test option_env match_none);
  assert (Test.CheckML.test option_env match_some);
  assert (Test.CheckML.test option_env match_some_annotated); *)
  (* we include some printing below because Dune
     appears to only show tests that show some output. *)
  print_endline "TestML: all tests passed."
